

# partially based on nokia file server code:
# Copyright (c) 2005 Nokia. All rights reserved.

import sys
import socket
import os
import binascii
import struct
import time
import StringIO
import shutil

if sys.platform != "win32": # enable test usage on pc
    import appuifw
    import e32
    import series60_console

import tc_protocol

CONFIG_DIR='%s:/system/apps/tcs60fs' % sys.argv[0][0]

DEBUGLOG=lambda x: sys.stdout.write(x+"\n")
#DEBUGLOG=lambda x: 0

class TCPacketError(Exception):
    pass

def file_checksums(filelist):
    retval={}
    for fname in filelist:
        try: 
            f=open(fname,'rb')
            checksum=binascii.crc32(f.read())
            f.close()
        except:
            checksum=None
        retval[fname]=checksum
    return retval

class fileserver:
    
    class TCState: # total commander protocol state
        overwrite = False
        restart_pos = 0        
        
    def __init__(self,sock):
        self.sock=sock
        self.commands={'quit': self.cmd_quit,
                       'get': self.cmd_get,
                       'put': self.cmd_put,
                       'exec': self.cmd_exec,
                       'eval': self.cmd_eval}
        
        self.tc_state = self.TCState()
        
    def cmd_quit(self,line):
        self.finished=True
    def cmd_invalid(self,line):
        print >>sys.stderr,'Invalid command '+line,
        self.finished=True        
    def cmd_get(self,cmdline):
        filename=eval(cmdline[4:])
        DEBUGLOG("get "+filename)
        f=open(filename,'rb')
        content=f.read()
        f.close()
        self.senddata(content)
    def cmd_put(self,cmdline):
        words=cmdline.split()
        filename=eval(words[1])
        DEBUGLOG("put "+filename)
        content=self.recvdata()
        f=open(filename,'wb')
        f.write(content)
        f.close()
        print "Wrote %d bytes to %s."%(len(content),filename)
    def senddata(self,data):
        DEBUGLOG("Content-Length: %d"%len(data))
        self.write('Content-Length: '+str(len(data))+'\n'+
                   'CRC32: '+str(binascii.crc32(data))+'\n')
        sentbytes=0        
        # Send the data in little bits because the Bluetooth serial
        # connection may lose data on large sends.
        MAX_SEND=2048
        while sentbytes<len(data):
            n=min(len(data),MAX_SEND)
            self.write(data[sentbytes:sentbytes+n])
            sentbytes+=n
        print "Sent %d bytes."%sentbytes
    def recvdata(self):
        DEBUGLOG("Waiting for data...")
        header_lines=[self.readline() for x in range(2)]
        header=dict([x.rstrip().split(': ') for x in header_lines])
        content_length=int(header['Content-Length'])
        crc32=int(header['CRC32'])
        DEBUGLOG("Content-Length: %d"%content_length)
        recvbytes=0
        content=[]
        DEBUGLOG("Receiving data...")
        while recvbytes<content_length:
            recvstring=self.sock.recv(min(content_length-recvbytes,2048))
            recvbytes+=len(recvstring)
            #print "Received: %d bytes (%3.1f%%)\r"%(recvbytes,(100.*recvbytes/content_length)),
            content.append(recvstring)
        print "Received %d bytes."%recvbytes
        content=''.join(content)
        if crc32 != binascii.crc32(content):
            print "*** CRC error!"        
        return content
    def cmd_exec(self,cmdline):
#        command=eval(cmdline[5:])
        command=eval(self.recvdata())
        DEBUGLOG("exec "+command)
        try:
            exec command in globals()
            result=(0,'')
        except:
            import traceback
            result=(1,apply(traceback.format_exception,sys.exc_info()))
        self.senddata(repr(result))

    def cmd_eval(self,cmdline):
        expr=eval(self.recvdata())
        #expr=eval(cmdline[5:])
        DEBUGLOG("eval "+expr)
        # two eval's because we need to first get rid of one level of
        # quoting.
        result=''
        try:
            value=eval(expr,globals())
            result=(0,value)
        except:
            import traceback
            result=(1,apply(traceback.format_exception,sys.exc_info()))
        self.senddata(repr(result))

    def readline(self):
        s=[]
        while 1:
            c=self.sock.recv(1)
            #self.sock.send(c)
            s.append(c)
            # We support just sane line-endings here.
            if c=='\n':
                break
        return ''.join(s)
    def write(self,msg):
        self.sock.send(msg)
    def run(self):
        self.finished=False
        while not self.finished:
            hdr = self.sock.recv(1)
            # check for total cmdr command
            if hdr == "F":
                self.process_totalcmdr_packet()     
            else:
                # process regular file server command
                cmdline=hdr+self.readline().rstrip()
                print "Received: "+cmdline
                words=cmdline.split()
                if len(words)>0:
                    cmd=words[0]
                    DEBUGLOG("Running command: "+cmdline)
                    self.commands.get(cmd,self.cmd_invalid)(cmdline)
                

    def process_totalcmdr_packet(self):
        # ASSUME that we have already read "F" from socket,
        # next stuff should be T and H

        try:
            magic, cmd, data = self.read_block(first_byte_read=True)
        except TCPacketError, msg:
            print >>sys.stderr, msg
            self.sock.send(tc_protocol.RQ_TRANSFERERR)
            return

        cmd_f = getattr(self, "tc_cmd_" + cmd, None)
        if not cmd_f:
            self.finished = True
            print >>sys.stderr, "invalid total commander command: %s" % cmd
            return

        try:
            cmd_f(data)
        finally:        
            # clear out latched overwrite or resume
            if cmd not in (tc_protocol.CMD_RESTART, tc_protocol.CMD_OVERWRITE):
                self.tc_state.restart_pos = 0
                self.tc_state.overwrite = False

    def tc_cmd_L(self, d):
        # list

        pth = self.check_and_mangle_path(d)
        if pth is None:
            print "path was invalid:", repr(d)
            self.sock.send(tc_protocol.RQ_FAILED)
            return

        self.sock.send(tc_protocol.RQ_OK)
        rpl = self.sock.recv(1)
        assert rpl == tc_protocol.RQ_READY

        #print pth
        if pth == "": # root path
            self.send_tc_block(tc_protocol.CMD_SIZE, str(0xFFFFFFFF))

            # XXX remove this when done testing
            if sys.platform == "win32":
                drives = [x.encode("latin1") + "\\" for x in (u"C:", u"D:")]
            else:
                drives = [x.encode("latin1") + "\\" for x in e32.drive_list()]

            block = '\r\n'.join(drives)
        else:
            files = os.listdir(pth)
            ffiles = []
            for fn in files:
                fpn = os.path.join(pth, fn)
                if os.path.isdir(fpn):
                    ffn = fn+"\\"
                else:
                    s = os.stat(fpn)
                    size = str(s[6])
                    tm = time.strftime("%Y%m%d%H%M%S", time.localtime(s[8]))
                    ffn = "\t".join((fn, size, tm))
                ffiles.append(ffn)

            block = "\r\n".join(ffiles)

        self.send_tc_body_block(StringIO.StringIO(block))
        self.send_tc_block(tc_protocol.CMD_EOF, '')

    
    def tc_cmd_G(self, d):
        pth = self.check_and_mangle_path(d)
        if pth is None or not os.path.isfile(pth):
            print "path was invalid:", repr(d)
            self.sock.send(tc_protocol.RQ_FAILED)
            return
        
        try:
            fh = open(pth, "rb")
        except:
            print "failed to open for read", pth
            self.sock.send(tc_protocol.RQ_FAILED)
            return

        print "sending", pth

        self.sock.send(tc_protocol.RQ_OK)
        rpl = self.sock.recv(1)
        assert rpl == tc_protocol.RQ_READY, rpl      
        
        self.send_tc_body_block(fh)
        self.send_tc_block(tc_protocol.CMD_EOF, '')
        fh.close()

    def tc_cmd_S(self, d):
        # send
        pth = self.mangle_path(d)

        if os.path.isfile(pth) and not (
            self.tc_state.restart_pos or self.tc_state.overwrite):
            #print "can't replace file if there is no overwrite nor resume set"
            # dont print error because the crazy protocol sends this anyway
            self.sock.send(tc_protocol.RQ_FAILED)
            return

        if (self.tc_state.restart_pos or self.tc_state.overwrite) and not \
               os.path.isfile(pth):
            print "resume or overwrite requested but file does not exists"
            self.sock.send(tc_protocol.RQ_FAILED)
            return

        try:
            if self.tc_state.restart_pos:
                s = os.stat(pth)
                size = s[6]
                if size != self.tc_state.restart_pos:
                    print "file resume position must match file size"
                    print "current size:", size, "restart pos:", \
                          self.tc_state.restart_pos
                    self.sock.send(tc_protocol.RQ_FAILED)
                    return
                    

            if self.tc_state.restart_pos:
                mode = "ab"
            else:
                mode = "wb"
                
            fh = open(pth, mode)
            orig_size = os.stat(pth)[6]
            
        except:
            print "opening file for writing failed"
            self.sock.send(tc_protocol.RQ_FAILED)
            return

        print "receiving", pth

        self.sock.send(tc_protocol.RQ_OK)
        self.sock.send(tc_protocol.RQ_READY)
        
        magic, cmd, data = self.read_block(first_byte_read=False)
        assert cmd == tc_protocol.CMD_SIZE
        totalsize = int(data.split("\t")[0])
        self.sock.send(tc_protocol.RQ_OK)

        if self.tc_state.restart_pos:
            totalsize -= orig_size
            # we might try to resume a file that has smaller size
            # than the new one. in that case, we do nothing.
            totalsize = max(0, totalsize)
            
        received = 0
        cnt = 0
        while received < totalsize:
            magic, cmd, data = self.read_block(first_byte_read=False)
            
            assert cmd == tc_protocol.CMD_BODY, cmd
            assert ord(magic) == cnt, (magic, cnt)

            received += len(data)
            
            fh.write(data)
            self.sock.send(tc_protocol.RQ_OK)
            
            if cnt == 255:
                cnt = 0
            else:
                cnt += 1
                
        fh.close()            
        assert received == totalsize
        
        magic, cmd, data = self.read_block(first_byte_read=False)
        assert cmd == tc_protocol.CMD_EOF, cmd
        self.sock.send(tc_protocol.RQ_OK)
        
    def tc_cmd_D(self, d):
        # delete
        pth = self.check_and_mangle_path(d)
        if pth is None:
            print "path was invalid:", pth
            self.sock.send(tc_protocol.RQ_FAILED)
            return

        print "deleting", pth
        
        try:
            os.unlink(pth)
        except:
            print "could not delete", pth
            self.sock.send(tc_protocol.RQ_FAILED)

        self.sock.send(tc_protocol.RQ_OK)

    def tc_cmd_F(self, d):
        # renfr
        pth = self.check_and_mangle_path(d)                
        if pth is None:
            print "rename source file must exist:", pth
            self.sock.send(tc_protocol.RQ_FAILED)
            return

        print pth
        self.sock.send(tc_protocol.RQ_OK)
        
        source_file = pth
        magic, cmd, data = self.read_block(first_byte_read=False)

        if cmd == tc_protocol.CMD_RENTO:            
            target_file = self.mangle_path(data)
            if os.path.exists(target_file):
                print "target folder must not exist"
                self.sock.send(tc_protocol.RQ_FAILED)
                return

            print "renaming", source_file, target_file

            try:
                os.rename(source_file, target_file)
            except:
                print "rename failed", source_file, target_file
                self.sock.send(tc_protocol.RQ_FAILED)
                return
                
            self.sock.send(tc_protocol.RQ_OK)
            
        elif cmd == tc_protocol.CMD_COPYTO:
            target_file = self.mangle_path(data)
            if os.path.exists(target_file):
                print "target must not exist when copying"
                self.sock.send(tc_protocol.RQ_FAILED)
                return

            print "copying", source_file, target_file

            try:
                self.sock.send(tc_protocol.RQ_WAIT)
                shutil.copyfile(source_file, target_file)
            except:
                print "copy failed", source_file, target_file
                self.sock.send(tc_protocol.RQ_FAILED)
                return

            self.sock.send(tc_protocol.RQ_OK)
            
        else:
            print "invalid commander after cmd_renfr", cmd
            self.sock.send(tc_protocol.RQ_FAILED)
            return
            
    def tc_cmd_M(self, d):
        # mkdir
        pth = self.mangle_path(d)

        print "makedir", pth
        
        try:
            os.mkdir(pth)
        except:
            print "could not create path", pth
            self.sock.send(tc_protocol.RQ_FAILED)
            return

        self.sock.send(tc_protocol.RQ_OK)


    def tc_cmd_C(self, d):
        # connect
        assert d == ""
        self.sock.send(tc_protocol.RQ_OK)

    def tc_cmd_A(self):
        # setattr
        self.sock.send(tc_protocol.RQ_FAILED)

    def tc_cmd_I(self):
        # settime
        self.sock.send(tc_protocol.RQ_FAILED)

    def tc_cmd_R(self, d):
        # restart
        self.tc_state.restart_pos = int(d[:-1])
        self.sock.send(tc_protocol.RQ_OK)

    def tc_cmd_Q(self, d):
        # getsize
        pth = self.check_and_mangle_path(d)
        if pth is None or not os.path.isfile(pth):
            print "path was invalid:", repr(d)
            self.sock.send(tc_protocol.RQ_FAILED)
            return

        self.sock.send(tc_protocol.RQ_OK)
        rpl = self.sock.recv(1)
        assert rpl == tc_protocol.RQ_READY, rpl
        
        size = str(os.stat(pth)[6])
        self.send_tc_block(tc_protocol.CMD_SIZE, size)

    def tc_cmd_W(self, d):
        # overwrite
        assert d == ""
        self.tc_state.overwrite = True
        self.tc_state.restart_pos = 0
        self.sock.send(tc_protocol.RQ_OK)

    def read_block(self, first_byte_read):
        if not first_byte_read:
            if self.sock.recv(1) != "F":
                raise TCPacketError("total commander packet must start "
                                    "with 'F'")
        hdr_begin = self.sock.recv(1) + self.sock.recv(1)
        cmd = self.sock.recv(1)

        if cmd != "b" and hdr_begin != "TH":
            raise TCPacketError("commands starting with F are " \
                  "reserved for total commander plugin")
        
        magic = hdr_begin[1] # for cmd_body (b) this is sequence number
        
        # read the rest of the basic block
        lenhi = self.sock.recv(1)
        lenlo = self.sock.recv(1)        
        block_len = (ord(lenhi) << 8) | ord(lenlo)

        hdr = "F" + hdr_begin + cmd + lenhi + lenlo
        c_crc = binascii.crc_hqx(hdr, 0)
        hdrcrc = self.sock.recv(1) + self.sock.recv(1)
        r_crc = struct.unpack(">H", hdrcrc)[0]

        if r_crc != c_crc:
            raise TCPacketError("crc error in total cmdr packet header")

        remaining = block_len - 8
        if remaining:
            data = []
            while remaining:
                buf = self.sock.recv(remaining)
                remaining -= len(buf)
                data.append(buf)
            
            data = ''.join(data)
            datacrcbytes = data[-2:]
            data = data[:-2]            

            r_crc = struct.unpack(">H", datacrcbytes)[0]
            c_crc = binascii.crc_hqx(data, 0)

            if r_crc != c_crc:
                raise TCPacketError("crc error in total cmdr packet payload")
        else:
            data = ''

        return magic, cmd, data


    def mangle_path(self, pth):
        pth = pth[1:-1] # remove \ at the start and zero at the end
        if pth and pth[-1] == ":": # a drive letter
            pth += "\\"
        return pth
    
    def check_and_mangle_path(self, pth):
        pth = self.mangle_path(pth)
        if not (pth == "" or              # root path
                os.path.exists(pth)): # file or directory
            return None
        return pth    

    def send_tc_body_block(self, data):
        # block should be StringIO object
        cnt = 0
        while 1:
            buf = data.read(tc_protocol.BLOCK_BUFFER)
            if not buf:
                break
            
            self.send_tc_block(tc_protocol.CMD_BODY, buf, cnt)
            
            if cnt == 255:
                cnt = 0
            else:
                cnt += 1

    def send_tc_block(self, cmd, block, seq=None):
        retry_count = 10

        while retry_count:
        
            datalen = 8 + len(block)
            if len(block):
                datalen += 2 # for crc

            if seq is not None:
                magic = chr(seq)
            else:
                magic = "H"

            # make header
            hdr1 = "FT%s%s%s" % (magic, cmd, struct.pack(">H", datalen))
            crc16 = binascii.crc_hqx(hdr1, 0)
            hdr = hdr1 + struct.pack(">H", crc16)

            # make payload
            if block:
                crc16 = binascii.crc_hqx(block, 0)
                payload = block + struct.pack(">H", crc16)
            else:
                payload = ''

            packet = hdr + payload
            self.sock.send(packet)
            rpl = self.sock.recv(1)

            if rpl == tc_protocol.RQ_CRCERR:
                print "crc error, retrying (retry count %d)" % retry_count
                retry_count -= 1
            elif rpl == tc_protocol.RQ_OK:
                break
            else:
                raise TCPacketError("reply was not ok: %d" % ord(rpl))
        else:
            assert "sending failed, too many crc errors"
            

def discover_address(configfile, dont_ask_if_found):
    import appuifw
    CONFIG_FILE=os.path.join(CONFIG_DIR,configfile)
    try:        
        f=open(CONFIG_FILE,'r')
        config=eval(f.read())
        f.close()
    except:
        config={}

    address=config.get('target','')

    if address and dont_ask_if_found:
        return address

    if address:
        choice=appuifw.popup_menu([u'Default host',
                                   u'Other...'],u'Connect to:')
        if choice==1:
            address=None
        if choice==None:
            return None # popup menu was cancelled.    
    if not address:
        print "Discovering..."
        addr,services=socket.bt_discover()
        print "Discovered: %s, %s"%(addr,services)
        if len(services)>1:
            choices=services.keys()
            choices.sort()
            choice=appuifw.popup_menu([unicode(services[x])+": "+x
                                       for x in choices],u'Choose port:')
            port=services[choices[choice]]
        else:
            port=services[services.keys()[0]]
        address=(addr,port)
        config['target']=address
        # make sure the configuration file exists.
        if not os.path.isdir(CONFIG_DIR):
            os.makedirs(CONFIG_DIR)
        f=open(CONFIG_FILE,'wt')
        f.write(repr(config))
        f.close()
    return address


# bad name, and bad design, but not very critical
class BTConnector:
    def __init__(self, use_default=False):
        self.use_default=use_default
    def runserver(self, connected_callback, finished_callback):
        addr = discover_address('fileserver_conf.txt', self.use_default)
        if not addr:
            return

        self.s=socket.socket(socket.AF_BT,socket.SOCK_STREAM)
        print "Connecting to "+`addr`
        self.s.connect(addr)
        server=fileserver(self.s)        

        connected_callback()

        try:
            server.run()
        finally:
            self.s.close()                
            finished_callback()
        
        print "Server finished."


    def shutdown(self):
        self.s.shutdown(2)
        self.s.close()

class USBConnector:    
    def runserver(self, connected_callback, finished_callback):
        # horrible hack to work-around the usb disconnection bug
        import pys60usb
        print "killing fax/modem process"
        pys60usb.kill_faxmodem()        
        
        already_reported = False
        print "Connecting with USB"
        try:
            while 1:                
                try:        
                    self.s = pys60usb.USBConnection()        
                    self.s.connect()
                    server=fileserver(self.s)

                    if not already_reported:
                        connected_callback()
                        already_reported=True

                    server.run()

                except SymbianError, err:
                    if err.errno == -29:
                        print "usb connexion down, trying again"
                        self.s.close()
                        del self.s
                        del server
                        continue
                    raise
            
        finally:
            self.s.close()                
            finished_callback()
        
        print "Server finished."

    def shutdown(self):
        self.s.close()

class SerialConnector:    
    # this is only for pc testing
    def runserver(self):
        import serial
        
        class SocketAdapter:
            def __init__(self, ser):
                self.ser=ser
            def recv(self,n=1):
                return self.ser.read(n)
            def send(self,msg):
                self.ser.write(msg)
            def close(self):
                self.ser.close()

        print "starting server"
        self.s=SocketAdapter(serial.Serial(3))
        
        server=fileserver(self.s)

        try:
            server.run()
        finally:
            self.s.close()

        print "Server finished."

    def shutdown(self):
        self.s.close()        


class App:
    def __init__(self):
        my_console = series60_console.Console()
        appuifw.app.body = my_console.text

        self.menu1 = [(u"Connect BT default", lambda : self.cmd_connect('bt_def')),
                      (u"Connect with BT", lambda : self.cmd_connect('bt')),                      
                      (u"Connect with USB", lambda : self.cmd_connect('usb')),
                      (u"Exit", self.cmd_quit)]
        
        self.menu2 = [(u"Disconnect", self.cmd_disconnect),
                      (u"Exit", self.cmd_quit)]
        appuifw.app.menu = self.menu1

        self.connector = None
        self.ui_lock = e32.Ao_lock()

    def run(self):        
        self.ui_lock.wait()

    def _connected_server_cb(self):
        appuifw.app.menu = self.menu2

    def _finished_server_cb(self):
        appuifw.app.menu = self.menu1

    def cmd_quit(self):
        self.ui_lock.signal()
    
    def cmd_connect(self, method):
        if method == 'usb':
            self.connector = USBConnector()
        elif method == "bt_def":
            self.connector = BTConnector(use_default=True)
        else:
            self.connector = BTConnector()
            
        run_it = lambda : self.connector.runserver(self._connected_server_cb,
                                                   self._finished_server_cb)

        # please contact me if you need explanation why this is done
        # this way. short reason: we use nested active scheduler to avoid
        # threading.
        e32.ao_sleep(0, run_it)

    def cmd_disconnect(self):
        self.connector.shutdown()
        
def main():
    app = App()
    app.run()

def test_in_pc():
    sc = SerialConnector()
    sc.runserver()
    
if __name__ == "__main__":
    if sys.platform == "win32":
        test_in_pc()
    else:
        main()
